#include <glib.h>
#include <glib/gprintf.h>
#include <glib-object.h>
#include <spice-session.h>
#include <spice-channel.h>
#include <channel-main.h>
#include <channel-cursor.h>
#include <channel-display.h>
#include <channel-inputs.h>
#include <SDL.h>


typedef struct sdl_context sdl_context;
struct sdl_context {
  gint width;
  gint height;
  gint stride;
  void *data;
  SDL_Window *win;
  SDL_Renderer *renderer;
  SDL_Texture *texture;
};

typedef struct spice_connection spice_connection;
struct spice_connection {
  SpiceSession     *session;
  SpiceMainChannel *main;
  SpiceDisplayChannel *display;
  SpiceInputsChannel *input;
  SpiceAudio       *audio;
  gchar            *name;
  sdl_context      sdl;
  int              channels;
  int              disconnecting;
};


static GMainLoop *mainloop = NULL;
spice_connection *conn;
static int connections = 0;


static void connection_destroy(void) {
  g_object_unref(conn->session);
  g_free(conn);

  connections--;
  if (connections > 0) {
    return;
  }

  g_main_loop_quit(mainloop);
}

static void connection_disconnect(void) {
  if (conn->disconnecting)
    return;
  conn->disconnecting = TRUE;
  spice_session_disconnect(conn->session);
}


static void main_channel_event(SpiceChannel *channel, SpiceChannelEvent event, gpointer data) {
  (void)data;
  (void)channel;
  char password[64];

  switch (event) {
    case SPICE_CHANNEL_OPENED:
      g_printf("main channel: opened\n");
      break;
    case SPICE_CHANNEL_SWITCHING:
      g_printf("main channel: switching host\n");
      break;
    case SPICE_CHANNEL_CLOSED:
      g_printf("main channel: closed\n");
      connection_disconnect();
      break;
    case SPICE_CHANNEL_ERROR_IO:
      connection_disconnect();
      break;
    case SPICE_CHANNEL_ERROR_TLS:
    case SPICE_CHANNEL_ERROR_LINK:
    case SPICE_CHANNEL_ERROR_CONNECT:
      g_printf("main channel: failed to connect\n");
      connection_disconnect();
      break;
    case SPICE_CHANNEL_ERROR_AUTH:
      g_printf("main channel: auth failure (wrong password?)\n");
      g_strlcpy(password, "", sizeof(password));
      connection_disconnect();
      break;
    default:
      g_printf("unknown main channel event: %d\n", event);
      break;
  }
}

static void main_mouse_update(SpiceChannel *channel) {
  gint mode;

  g_object_get(channel, "mouse-mode", &mode, NULL);
  switch (mode) {
    case SPICE_MOUSE_MODE_SERVER:
      g_printf("mouse mode: server\n"); 
      break;
    case SPICE_MOUSE_MODE_CLIENT:
      g_printf("mouse mode: client\n"); 
      break;
  }
}





static void display_monitors(SpiceChannel *display, GParamSpec *pspec, gpointer data) {
  GArray *monitors = NULL;
  (void)pspec;
  (void)data;
  int id;

  g_object_get(display, "channel-id", &id, "monitors", &monitors, NULL);
  g_return_if_fail(monitors != NULL);
  g_printf("display_monitors id:%d\n", id);

  if(conn->sdl.width > 0 && conn->sdl.height > 0) {
    conn->sdl.win = SDL_CreateWindow("spice sdl", SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED, conn->sdl.width, conn->sdl.height, SDL_WINDOW_OPENGL);
    if(conn->sdl.win == NULL) {
      g_error("unable to create window\n%s\n", SDL_GetError());
    }
    conn->sdl.renderer = SDL_CreateRenderer(conn->sdl.win, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_TARGETTEXTURE);
    if(conn->sdl.renderer == NULL) {
      g_error("unable to create rederer\n%s\n", SDL_GetError());
    }
    conn->sdl.texture = SDL_CreateTexture(conn->sdl.renderer, SDL_PIXELFORMAT_ARGB8888, SDL_TEXTUREACCESS_STATIC, conn->sdl.width, conn->sdl.height);
    if(conn->sdl.texture == NULL) {
      g_error("unable to create texture\n%s\n", SDL_GetError());
    }
    SDL_RenderClear(conn->sdl.renderer);
  } else {
    g_printf("cannot open window as there is no dimesion stored\n");
  }
  g_clear_pointer(&monitors, g_array_unref);
}

/*
 * the gpointer sdl_surf is here present, because I tried to make some changes on the channel as well. you may remove this argument and make the argument list like:
 * (SpiceChannel *channel, gint format, gint width, gint height, gint stride, gint shmid, gpointer imgdata, gpointer data)
 */
static void primary_create(SpiceChannel *channel, gint format, gint width, gint height, gint stride, gint shmid, gpointer sdl_surf, gpointer imgdata, gpointer data) {
  (void)channel;
  (void)shmid;
  (void)sdl_surf;
  (void)data;
  conn->sdl.width = width;
  conn->sdl.height = height;
  conn->sdl.stride = stride;
  conn->sdl.data = imgdata;
  g_printf("creating primary: %dx%d, format %d, stride: %d\n", width, height, format, stride);

}

static void invalidate(SpiceChannel *channel, gint x, gint y, gint w, gint h, gpointer *data) {
  (void)channel;
  (void)data;
  SDL_Rect s_rect;

  if(conn->sdl.renderer && conn->sdl.texture && conn->sdl.data && conn->sdl.stride) {

    s_rect.x = x;
    s_rect.y = y;
    s_rect.w = w;
    s_rect.h = h;
    SDL_UpdateTexture(conn->sdl.texture, NULL, conn->sdl.data, conn->sdl.stride);
    SDL_RenderCopy(conn->sdl.renderer, conn->sdl.texture, NULL, NULL);
    SDL_RenderPresent(conn->sdl.renderer);
  } else {
    g_printf("surface is not created yet\n");
  }
}


static void channel_new(SpiceSession *s, SpiceChannel *channel, gpointer data) {
  (void)data;
  int id;

  g_object_get(channel, "channel-id", &id, NULL);
  conn->channels++;
  g_printf("new channel (#%d)\n", id); 

  if (SPICE_IS_MAIN_CHANNEL(channel)) {
    g_printf("new main channel\n");
    conn->main = SPICE_MAIN_CHANNEL(channel);
    g_signal_connect(channel, "channel-event", G_CALLBACK(main_channel_event), NULL);
    g_signal_connect(channel, "main-mouse-update", G_CALLBACK(main_mouse_update), NULL);
    main_mouse_update(channel);
  }
  if (SPICE_IS_DISPLAY_CHANNEL(channel)) {
    g_printf("new display channel (#%d)\n", id); 
    g_signal_connect(channel, "display-primary-create", G_CALLBACK(primary_create), NULL);
    g_signal_connect(channel, "notify::monitors", G_CALLBACK(display_monitors), NULL);
    g_signal_connect(channel, "display-invalidate", G_CALLBACK(invalidate), NULL);
    conn->display = SPICE_DISPLAY_CHANNEL(channel);
    spice_channel_connect(channel);
  }
  if (SPICE_IS_INPUTS_CHANNEL(channel)) {
    g_printf("new inputs channel\n");
    //g_signal_connect(channel, "inputs-modifiers", G_CALLBACK(inputs_modifiers), NULL);
    conn->input = SPICE_INPUTS_CHANNEL(channel);
    spice_channel_connect(channel);
  }
  if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
    g_printf("new audio channel\n");
    conn->audio = spice_audio_get(s, NULL);
  }
}



static void channel_destroy(SpiceSession *s, SpiceChannel *channel, gpointer data) {
  (void)data;
  (void)s;
  int id;

  g_object_get(channel, "channel-id", &id, NULL);
  if (SPICE_IS_MAIN_CHANNEL(channel)) {
    g_printf("zap main channel\n");
    conn->main = NULL;
  }

  if (SPICE_IS_DISPLAY_CHANNEL(channel)) {
    g_printf("zap display channel (#%d)\n", id);
  }

  if (SPICE_IS_PLAYBACK_CHANNEL(channel)) {
    g_printf("zap audio channel\n");
  }

  conn->channels--;
  if (conn->channels > 0) {
    return;
  }

  connection_destroy();
}


static spice_connection *connection_new(void) {
  conn = g_new0(spice_connection, 1);
  conn->session = spice_session_new();
  g_signal_connect(conn->session, "channel-new", G_CALLBACK(channel_new), NULL);
  g_signal_connect(conn->session, "channel-destroy", G_CALLBACK(channel_destroy), NULL);
  //g_signal_connect(conn->session, "notify::migration-state", G_CALLBACK(migration_state), NULL);
  connections++;
  return conn;
}


static void connection_connect(void) {
  conn->disconnecting = FALSE;
  spice_session_connect(conn->session);
}




gint sdl_to_spice_buttonmask(gint sdl_mask) {
  gint spice_mask = 0;
  if (sdl_mask & SDL_BUTTON_LMASK)
    spice_mask |= SPICE_MOUSE_BUTTON_MASK_LEFT;
  if (sdl_mask & SDL_BUTTON_MMASK)
    spice_mask |= SPICE_MOUSE_BUTTON_MASK_MIDDLE;
  if (sdl_mask & SDL_BUTTON_RMASK)
    spice_mask |= SPICE_MOUSE_BUTTON_MASK_RIGHT;
  return(spice_mask);
}


gboolean check_events(void *data) {
  (void)data;
  SDL_Event event;
  gint button_mask, x, y, id = 0;

  if(SDL_PollEvent(&event)) {
    switch(event.type) {
      case SDL_QUIT:
	g_main_loop_quit(mainloop);
	break;
      case SDL_MOUSEMOTION:
      case SDL_MOUSEBUTTONDOWN:
      case SDL_MOUSEBUTTONUP:
	button_mask = SDL_GetMouseState(&x, &y);
	spice_inputs_position(conn->input, x, y, id, sdl_to_spice_buttonmask(button_mask));
	break;
    }
  }

  return(TRUE);
}



int main(int argc, char **argv) {
  gint i = 1;
  gchar *host = NULL, *port = NULL, *pwd = NULL;

  mainloop = g_main_loop_new(NULL, FALSE);

  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_EVENTS) != 0) {
    g_error("unable to initialize SDL");
  }

  conn = connection_new();
  while(i < argc) {
    if(argv[i][0] == '-' && argv[i][1] == 'h') {
      /* -h is for the host */
      i++;
      if(argc > i) {
	host = strdup(argv[i]);
      }
    } else if(argv[i][0] == '-' && argv[i][1] == 'p') {
      /* -p is for the port */
      i++;
      if(argc > i) {
	port = strdup(argv[i]);
      }
    } else if(argv[i][0] == '-' && argv[i][1] == 'w') {
      /* -w is for the password */
      i++;
      if(argc > i) {
	pwd = strdup(argv[i]);
      }
    } else if(argv[i][0] == '-' && argv[i][1] == 'n') {
      /* -n is for the name */
      i++;
      if(argc > i) {
	conn->name = strdup(argv[i]);
      }
    }
    i++;
  }

  if(!host || !port) {
    g_printf("host (-h) and port (-p) are required\n");
    return(1);
  }

  spice_set_session_option(conn->session);

  g_object_set(conn->session, "host", host, NULL);
  g_object_set(conn->session, "port", port, NULL);
  if(pwd) {
    g_object_set(conn->session, "password", pwd, NULL);
  }

  g_timeout_add(100, check_events, NULL);

  connection_connect();
  if(connections > 0) {
    g_main_loop_run(mainloop);
  }

  g_main_loop_unref(mainloop);
  SDL_Quit();
  return(0);
}

